<?php

declare(strict_types = 1);

namespace Drupal\rdf_entity;

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Config\Entity\ConfigEntityStorageInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\rdf_entity\Entity\RdfEntityMapping;

/**
 * Contains helper methods for managing the RDF entity graphs.
 */
class RdfGraphHandler implements RdfGraphHandlerInterface {

  /**
   * Static cache.
   *
   * @var array
   */
  protected $cache = self::EMPTY_CACHE;

  /**
   * The entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The RDF entity graph config entity storage.
   *
   * @var \Drupal\Core\Config\Entity\ConfigEntityStorageInterface
   */
  protected $rdfEntityGraphStorage;

  /**
   * Constructs a RDF graph handler object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager service.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager) {
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * {@inheritdoc}
   */
  public function getGraphDefinitions(string $entity_type_id): array {
    if (!isset($this->cache['definition'][$entity_type_id])) {
      $query = $this->getRdfEntityGraphStorage()->getQuery();
      $ids = $query->condition($query->orConditionGroup()
        ->condition('entity_types.*', $entity_type_id)
        // A NULL value means "all entity types".
        ->notExists('entity_types')
      )->condition('status', TRUE)
        // A determined order is a key feature.
        ->sort('weight', 'ASC')
        ->execute();

      if (!$ids) {
        // Do not cache an empty set, it may occur because this runs before any
        // configuration has been imported, so the entities are not yet in.
        return [];
      }

      $graphs = $this->getRdfEntityGraphStorage()->loadMultiple($ids);

      $this->cache['definition'][$entity_type_id] = array_map(function (RdfEntityGraphInterface $graph): array {
        return [
          'title' => $graph->label(),
          'description' => $graph->getDescription(),
        ];
      }, $graphs);
    }
    return $this->cache['definition'][$entity_type_id];
  }

  /**
   * {@inheritdoc}
   */
  public function getEntityTypeGraphIds(string $entity_type_id, array $limit_to_graph_ids = NULL): array {
    $graph_ids = array_keys($this->getGraphDefinitions($entity_type_id));
    if ($limit_to_graph_ids) {
      $graph_ids = array_intersect($graph_ids, $limit_to_graph_ids);
    }
    return $graph_ids;
  }

  /**
   * {@inheritdoc}
   */
  public function getDefaultGraphId(string $entity_type_id): string {
    $graph_ids = $this->getEntityTypeGraphIds($entity_type_id);
    return reset($graph_ids);
  }

  /**
   * {@inheritdoc}
   */
  public function getBundleGraphUri(string $entity_type_id, string $bundle, string $graph_id): ?string {
    $entity_type = $this->entityTypeManager->getDefinition($entity_type_id);
    $bundle_key = ($entity_type->hasKey('bundle') && $entity_type->getBundleEntityType()) ? $bundle : $entity_type_id;
    return $this->getEntityTypeGraphUris($entity_type_id)[$bundle_key][$graph_id] ?? NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function getEntityTypeGraphUris(string $entity_type_id, array $limit_to_graph_ids = NULL): array {
    if (!isset($this->cache['structure'][$entity_type_id])) {
      $entity_type = $this->entityTypeManager->getDefinition($entity_type_id);
      if ($entity_type->hasKey('bundle') && ($bundle_entity_id = $entity_type->getBundleEntityType())) {
        $bundle_keys = array_values($this->entityTypeManager->getStorage($bundle_entity_id)->getQuery()->execute());
      }
      else {
        $bundle_keys = [$entity_type_id];
      }

      foreach ($bundle_keys as $bundle_key) {
        $graphs = ($mapping = RdfEntityMapping::loadByName($entity_type_id, $bundle_key)) ? $mapping->getGraphs() : [];
        $this->cache['structure'][$entity_type_id][$bundle_key] = $graphs;
      }
    }

    // Limit the results.
    if ($limit_to_graph_ids) {
      return array_map(function (array $graphs) use ($limit_to_graph_ids): array {
        return array_intersect_key($graphs, array_flip($limit_to_graph_ids));
      }, $this->cache['structure'][$entity_type_id]);
    }

    return $this->cache['structure'][$entity_type_id];
  }

  /**
   * {@inheritdoc}
   */
  public function bundleHasGraph(string $entity_type_id, string $bundle, string $graph_id): bool {
    $entity_type_graphs = $this->getEntityTypeGraphUris($entity_type_id);
    return !empty($entity_type_graphs[$bundle][$graph_id]);
  }

  /**
   * {@inheritdoc}
   */
  public function getEntityTypeGraphUrisFlatList(string $entity_type_id, array $limit_to_graph_ids = NULL): array {
    $graphs = $this->getEntityTypeGraphUris($entity_type_id, $limit_to_graph_ids);
    return array_reduce($graphs, function (array $uris, array $bundle_graphs): array {
      return array_merge($uris, array_values($bundle_graphs));
    }, []);
  }

  /**
   * {@inheritdoc}
   */
  public function getBundleGraphId(string $entity_type_id, string $bundle, string $graph_uri): ?string {
    $graphs = $this->getEntityTypeGraphUris($entity_type_id);
    $search = array_search($graph_uri, $graphs[$bundle]);
    return $search !== FALSE ? $search : NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function clearCache(array $path = NULL): void {
    if (empty($path)) {
      $this->cache = static::EMPTY_CACHE;
      return;
    }
    NestedArray::unsetValue($this->cache, $path);

    // If the path was a top-level cache category, restore its "empty version".
    if (count($path) === 1 && array_key_exists($path[0], static::EMPTY_CACHE)) {
      $this->cache[$path[0]] = static::EMPTY_CACHE[$path[0]];
    }
  }

  /**
   * Returns the RDF entity graph config entity storage service.
   *
   * @return \Drupal\Core\Config\Entity\ConfigEntityStorageInterface
   *   The RDF entity graph config entity storage service.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   *   If the RDF entity type is not found.
   */
  protected function getRdfEntityGraphStorage(): ConfigEntityStorageInterface {
    if (!isset($this->rdfEntityGraphStorage)) {
      $this->rdfEntityGraphStorage = $this->entityTypeManager->getStorage('rdf_entity_graph');
    }
    return $this->rdfEntityGraphStorage;
  }

}
